@@ -1,18 +1,28 @@
+use std::io::Result;
 use std::os::unix::fs::{FileExt, FileTypeExt};
-use std::os::unix::io::AsRawFd;
+use std::os::unix::io::{AsRawFd, RawFd};
 use std::process::exit;
 
 use libc::{c_ushort, c_int};
 use nix::fcntl::{posix_fadvise, PosixFadviseAdvice};
 use nix::{ioctl_read, ioctl_read_bad, request_code_none};
 
-const MIB: u64 = 1024 * 1024;
-
 ioctl_read_bad!(blksectget, request_code_none!(0x12, 103), c_ushort);
 ioctl_read_bad!(blksszget, request_code_none!(0x12, 104), c_int);
 ioctl_read!(blkgetsize64, 0x12, 114, u64);
 
-fn main() -> std::io::Result<()> {
+fn willneed(fd: RawFd, offset: u64, len: u64) -> Result<()> {
+	for advice in [
+		PosixFadviseAdvice::POSIX_FADV_SEQUENTIAL,
+		PosixFadviseAdvice::POSIX_FADV_NOREUSE,
+		PosixFadviseAdvice::POSIX_FADV_WILLNEED] {
+		posix_fadvise(fd, offset.try_into().unwrap(), len.try_into().unwrap(), advice).map(|_| ())?;
+	}
+
+	Ok(())
+}
+
+fn main() -> Result<()> {
 	let mut args = std::env::args();
 
 	if args.len() != 2 {
@@ -80,39 +90,52 @@ fn main() -> std::io::Result<()> {
 	};
 
 	// Assert that device size is a multiple of the logical sector size
-	assert_eq!(size % ssz as u64, 0);
+	assert!(size % ssz as u64 == 0);
 
 	let null = vec![0u8; ssz];
 	let mut buffer = vec![0u8; sect * ssz];
 
-	for advice in [
-		PosixFadviseAdvice::POSIX_FADV_SEQUENTIAL,
-		PosixFadviseAdvice::POSIX_FADV_NOREUSE,
-		PosixFadviseAdvice::POSIX_FADV_WILLNEED] {
-		posix_fadvise(dev.as_raw_fd(), 0, 0, advice).unwrap_or_else(|err| {
-			eprintln!("Failed to predeclare access pattern for {}: {}", path, err);
-			exit(74);
-		});
-	}
+	willneed(dev.as_raw_fd(), 0, 0).unwrap_or_else(|err| {
+		eprintln!("Failed to predeclare access pattern for {}: {}", path, err);
+		exit(74);
+	});
 
 	let mut offset = 0u64;
 	let mut verify = 0usize;
+	let mut sync;
 
 	println!();
 
 	loop {
-		eprintln!("\x1bM{} / {} MiB\t({} %)", offset / MIB, size / MIB, offset * 100 / size);
+		if verify == 0 {
+			eprintln!("\x1bM\x1b[K{:>3} %  {:>11} / {}", offset * 100 / size,
+			          bytesize::to_string(offset, true), bytesize::to_string(size, true));
+		}
+
+		// Synchronise to device if last sector of range
+		sync = verify == 1;
+
 		match dev.read_at(if verify == 0 { &mut buffer } else { &mut buffer[0..ssz] }, offset) {
 			Ok(0) => { break; }
 			Ok(len) => {
-				assert_eq!(len % ssz, 0);
+				// Assert that we read a multiple of the sector size
+				assert!(len % ssz == 0);
+
 				offset += len as u64;
 				verify = verify.saturating_sub(1);
 			}
+
 			Err(err) => {
 				if let Some(libc::EIO) = err.raw_os_error() {
 					if verify == 0 {
 						verify = sect;
+
+						// Declare our intention to re‐read the range
+						willneed(dev.as_raw_fd(), offset, (sect * ssz) as u64).unwrap_or_else(|err| {
+							eprintln!("Failed to predeclare access pattern for range: {}", err);
+							exit(74);
+						});
+
 						continue;
 					}
 
@@ -122,6 +145,7 @@ fn main() -> std::io::Result<()> {
 						exit(74);
 					});
 
+					// Assert that we wrote a complete sector
 					assert_eq!(len, ssz);
 
 					offset += ssz as u64;
@@ -132,6 +156,13 @@ fn main() -> std::io::Result<()> {
 				}
 			}
 		}
+
+		if sync {
+			dev.sync_data().unwrap_or_else(|err| {
+				eprintln!("Failed to synchronise data to device: {}", err);
+				exit(74);
+			});
+		}
 	}
 
 	Ok(())
